import sys
sys.path.append('..')

import unittest
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

from mixins import SaveableMixin

class Derived(SaveableMixin):
    _PICKLER = 'pickle'

    def __init__(self, a: int = -1, b: str = 'unset', **kwargs):
        super().__init__(**kwargs)
        self.a = a
        self.b = b

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and (self.a == other.a) and (self.b == other.b)

class DillDerived(SaveableMixin):
    _PICKLER = 'dill'

    def __init__(self, a: int = -1, b: str = 'unset', **kwargs):
        super().__init__(**kwargs)
        self.a = a
        self.b = b

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and (self.a == other.a) and (self.b == other.b)

class BadDerived(SaveableMixin):
    _PICKLER = 'not_supported'

    def __init__(self, a: int = -1, b: str = 'unset', **kwargs):
        super().__init__(**kwargs)
        self.a = a
        self.b = b

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and (self.a == other.a) and (self.b == other.b)

class TestSaveableMixin(unittest.TestCase):
    def test_saveable_mixin(self):
        T = Derived(a=2, b='hi')

        with TemporaryDirectory() as d:
            save_path = Path(d) / 'save.pkl'
            T._save(save_path)

            with self.assertRaises(FileExistsError):
                new_t = Derived(a=3, b='bar')
                new_t._save(save_path)

            got_T = Derived._load(save_path)
            self.assertEqual(T, got_T)

            bad_T = BadDerived(a=2, b='hi')
            with self.assertRaises(NotImplementedError): bad_T._save(Path(d) / 'no_save.pkl')

            # This should error as that pickler isn't supported.
            with self.assertRaises(FileNotFoundError): got_T = Derived._load(Path(d) / 'no_save.pkl')
            with self.assertRaises(IsADirectoryError): got_T = Derived._load(Path(d))

            # This should error as dill isn't installed.
            with self.assertRaises(ImportError): bad_T = DillDerived(a=3, b='baz')
            T._PICKLER = 'dill'
            with self.assertRaises(ImportError): T._save(Path(d) / 'no_save.pkl')
            Derived._PICKLER = 'dill'
            with self.assertRaises(ImportError): got_T = Derived._load(save_path)

if __name__ == '__main__': unittest.main()



from __future__ import annotations

import functools, random, numpy as np

from datetime import datetime
from typing import Optional

from .utils import doublewrap

def seed_everything(seed: Optional[int] = None, try_import_torch: Optional[bool] = True) -> int:
    max_seed_value = np.iinfo(np.uint32).max
    min_seed_value = np.iinfo(np.uint32).min

    try:
        if seed is None: seed = os.environ.get("PL_GLOBAL_SEED")
        seed = int(seed)
    except (TypeError, ValueError):
        seed = np.random.randint(min_seed_value, max_seed_value)

    assert (min_seed_value <= seed <= max_seed_value)

    random.seed(seed)
    np.random.seed(seed)
    if try_import_torch:
        try:
            import torch
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
        except ModuleNotFoundError: pass

    return seed

class SeedableMixin():
    def __init__(self, *args, **kwargs):
        self._past_seeds = kwargs.get('_past_seeds', [])

    def _last_seed(self, key: str):
        for idx, (s, k, time) in enumerate(self._past_seeds[::-1]):
            if k == key:
                idx = len(self._past_seeds) - 1 - idx
                return idx, s

        return -1, None

    def _seed(self, seed: Optional[int] = None, key: Optional[str] = None):
        if seed is None: seed = random.randint(0, int(1e8))
        if key is None: key = ''
        time = str(datetime.now())

        self.seed = seed
        if hasattr(self, '_past_seeds'): self._past_seeds.append((self.seed, key, time))
        else: self._past_seeds = [(self.seed, key, time)]

        seed_everything(seed)
        return seed

    @staticmethod
    @doublewrap
    def WithSeed(fn, key: Optional[str] = None):
        if key is None: key = fn.__name__
        @functools.wraps(fn)
        def wrapper_seeding(self, *args, seed: Optional[int] = None, **kwargs):
            self._seed(seed=seed, key=key)
            return fn(self, *args, **kwargs)
        return wrapper_seeding




from __future__ import annotations

import functools, inspect, pickle
from copy import deepcopy
from pathlib import Path
from typing import Optional

from .utils import doublewrap

class DebuggableMixin:
    @property
    def _do_debug(self):
        if hasattr(self, 'do_debug'): return self.do_debug
        else: return False

    @staticmethod
    @doublewrap
    def CaptureErrorState(fn, store_global: Optional[bool] = None, filepath: Optional[Path] = None):
        if store_global is None: store_global = (filepath is None)

        @functools.wraps(fn)
        def debugging_wrapper(self, *args, seed: Optional[int] = None, **kwargs):
            if not self._do_debug: return fn(self, *args, **kwargs)

            try:
                return fn(self, *args, **kwargs)
            except Exception as e:
                T = inspect.trace()
                for t in T:
                    if t[3] == fn.__name__: break

                new_vars = deepcopy(t[0].f_locals)
                if store_global:
                    __builtins__["_DEBUGGER_VARS"] = new_vars
                if filepath:
                    with open(filepath, mode='wb') as f:
                        pickle.dump(new_vars, f)
                raise
        return debugging_wrapper



import sys
sys.path.append('..')

import unittest

from mixins import SwapcacheableMixin

class TestSwapcacheableMixin(unittest.TestCase):
    def test_constructs(self):
        T = SwapcacheableMixin()

if __name__ == '__main__': unittest.main()




from __future__ import annotations

import time
from typing import Hashable

class SwapcacheableMixin():
    def __init__(self, *args, **kwargs):
        self._cache_size = kwargs.get('cache_size', 5)

    def _init_attrs(self):
        if not hasattr(self, '_cache'): self._cache = {'keys': [], 'values': []}
        if not hasattr(self, '_cache_size'): self._cache_size = 5
        if not hasattr(self, '_front_attrs'): self._front_attrs = []
        if not hasattr(self, '_front_cache_key'): self._front_cache_key = None
        if not hasattr(self, '_front_cache_idx'): self._front_cache_idx = None

    def _set_swapcache_key(self, key: Hashable) -> None:
        self._init_attrs()
        if key == self._front_cache_key: return

        seen_key = self._swapcache_has_key(key)
        if seen_key:
            idx = next(i for i, (k, t) in enumerate(self._seen_parameters) if k == key)
        else:
            self._cache['keys'].append((key, time.time()))
            self._cache['values'].append({})

            self._cache['keys'] = self._cache['keys'][-self._cache_size:]
            self._cache['values'] = self._cache['values'][-self._cache_size:]

            idx = -1

        # Clear out the old front-and-center attributes
        for attr in self._front_attrs: delattr(self, attr)

        self._front_cache_key = key
        self._front_cache_idx = idx

        self._update_front_attrs()

    def _swapcache_has_key(self, key: Hashable) -> bool:
        self._init_attrs()
        return any(k == key for k, t in self._cache['keys'])

    def _swap_to_key(self, key: Hashable) -> None:
        self._init_attrs()
        assert self._swapcache_has_key(key)
        self._set_swapcache_key(key)

    def _update_front_attrs(self):
        self._init_attrs()
        # Set the new front-and-center attributes
        for key, val in self._cache['values'][self._front_cache_idx].items(): setattr(self, key, val)

    def _update_swapcache_key_and_swap(self, key: Hashable, values_dict: dict):
        self._init_attrs()
        assert key is not None

        self._set_swapcache_key(key)
        self._cache['values'][self._front_cache_idx].update(values_dict)
        self._update_front_attrs()

    def _update_current_swapcache_key(self, values_dict: dict):
        self._init_attrs()
        self._update_swapcache_key_and_swap(self._front_cache_key, values_dict)



# __all__ = [
#     'debuggable',
#     'multiprocessingable',
#     'saveable',
#     'swapcacheable',
#     'tensorable',
#     'timeable',
#     'tqdmable',
# ]

from .debuggable import DebuggableMixin
from .multiprocessingable import MultiprocessingMixin
from .saveable import SaveableMixin
from .seedable import SeedableMixin
from .swapcacheable import SwapcacheableMixin
from .timeable import TimeableMixin

# Tensorable and Tqdmable rely on packages that may or may not be installed.

try: from .tensorable import TensorableMixin
except ImportError as e: pass

try: from .tqdmable import TQDMableMixin
except ImportError as e: pass



import sys
sys.path.append('..')

import unittest

import random, numpy as np

from mixins import SeedableMixin

class SeedableDerived(SeedableMixin):
    def __init__(self):
        self.foo = 'foo'
        # Doesn't call super().__init__()! Should still work in this case.

    def gen_random_num(self):
        # This is purely random
        return (random.random(), np.random.rand())

    @SeedableMixin.WithSeed(key="decorated")
    def decorated_gen_random_num(self):
        return (random.random(), np.random.rand())

    @SeedableMixin.WithSeed
    def decorated_auto_key(self):
        return random.random()

class TestSeedableMixin(unittest.TestCase):
    def test_constructs(self):
        T = SeedableMixin()
        T = SeedableDerived()

    def test_responds_to_methods(self):
        T = SeedableMixin()

        T._seed()
        T._last_seed('foo')

        T = SeedableDerived()
        T._seed()
        T._last_seed('foo')


    def test_seeding_freezes_randomness(self):
        T = SeedableDerived()

        unseeded_1 = T.gen_random_num()
        unseeded_2 = T.gen_random_num()

        # Without seeding, repeated calls should be different.
        self.assertNotEqual(unseeded_1, unseeded_2)

        T._seed(1)
        seeded_1_1 = T.gen_random_num()
        seeded_2_1 = T.gen_random_num()

        # Even if I seeded at the start, repeated calls should still be different.
        self.assertNotEqual(seeded_1_1, seeded_2_1)

        T._seed(1)
        seeded_1_2 = T.gen_random_num()
        seeded_2_2 = T.gen_random_num()

        # Since I seeded again, they should match the prior sequence.
        self.assertEqual(seeded_1_1, seeded_1_2)
        self.assertEqual(seeded_2_1, seeded_2_2)

    def test_decorated_seeding_freezes_randomness(self):
        T = SeedableDerived()

        unseeded_1 = T.decorated_gen_random_num()
        unseeded_2 = T.decorated_gen_random_num()

        # Without seeding, repeated calls should be different.
        self.assertNotEqual(unseeded_1, unseeded_2)

        seeded_1_1 = T.decorated_gen_random_num(seed=1)
        seeded_2_1 = T.decorated_gen_random_num(seed=2)

        # Even if I seeded at the start, repeated calls should still be different.
        self.assertNotEqual(seeded_1_1, seeded_2_1)

        seeded_1_2 = T.decorated_gen_random_num(seed=1)
        seeded_2_2 = T.decorated_gen_random_num(seed=2)

        # Since they are seeded, they should match the prior sequence.
        self.assertEqual(seeded_1_1, seeded_1_2)
        self.assertEqual(seeded_2_1, seeded_2_2)

        # Now we want to make sure the seeding is consistent even interrupted.

        T._seed(0)
        seeded_1_3 = T.decorated_gen_random_num(seed=1)
        T._seed(10)
        seeded_2_3 = T.decorated_gen_random_num(seed=2)

        self.assertEqual(seeded_1_1, seeded_1_3)
        self.assertEqual(seeded_2_1, seeded_2_3)


    def test_seeds_follow_consistent_sequence(self):
        T = SeedableDerived()

        unseeded_seq = [T._seed() for i in range(5)]

        seed_1 = T._seed(1)

        # seed_1 should be 1 given I passed a seed in:
        self.assertEqual(seed_1, 1)

        next_seeds_1 = [T._seed() for i in range(5)]

        # These should differ from the unseeded sequence of seeds
        self.assertNotEqual(unseeded_seq, next_seeds_1)

        T._seed(1)

        next_seeds_2 = [T._seed() for i in range(5)]

        # The sequence of seeds should be the same here.
        self.assertEqual(next_seeds_1, next_seeds_2)

    def test_get_last_seed(self):
        T = SeedableDerived()

        key = 'key'
        non_key = 'not_key'

        seed_key_early = 1
        seed_key_late  = 1
        seed_non_key   = 2

        T._seed()

        idx, seed = T._last_seed(key)
        self.assertEqual(idx, -1)
        self.assertEqual(seed, None)

        T._seed(seed_key_early, key)
        T._seed()
        T._seed(seed_non_key, non_key)
        T._seed(seed_key_late, key)
        T._seed(seed_non_key, non_key)

        idx, seed = T._last_seed(key)
        self.assertEqual(idx, 4)
        self.assertEqual(seed, seed_key_late)

if __name__ == '__main__': unittest.main()




from __future__ import annotations

import pickle as pickle
try:
    import dill
    dill_imported = True
    dill_import_error = None
except ImportError as e:
    dill_import_error = e
    dill_imported = False

from pathlib import Path
from typing import Optional

class SaveableMixin():
    _DEL_BEFORE_SAVING_ATTRS = []
    # TODO(mmd): Make StrEnum upon conversion to python 3.11
    _PICKLER = 'dill' if dill_imported else 'pickle'

    def __init__(self, *args, **kwargs):
        self.do_overwrite = kwargs.get('do_overwrite', False)
        if self._PICKLER == 'dill' and not dill_imported: raise dill_import_error

    @classmethod
    def _load(cls, filepath: Path, **add_kwargs) -> None:
        if not filepath.exists():
            raise FileNotFoundError(f"{filepath} does not exist.")
        elif not filepath.is_file():
            raise IsADirectoryError(f"{filepath} is not a file.")

        with open(filepath, mode='rb') as f:
            match cls._PICKLER:
                case 'dill':
                    if not dill_imported: raise dill_import_error
                    obj = dill.load(f)
                case 'pickle': obj = pickle.load(f)
                case _:
                    raise NotImplementedError(f"{cls._PICKLER} not supported! Options: {'dill', 'pickle'}")

        for a, v in add_kwargs.items(): setattr(obj, a, v)
        obj._post_load(add_kwargs)

        return obj

    def _post_load(self, load_add_kwargs: dict) -> None:
        # Overwrite this in the base class if desired.
        return

    def _save(self, filepath: Path, do_overwrite: Optional[bool] = False) -> None:
        if not hasattr(self, 'do_overwrite'): self.do_overwrite = False
        if not (self.do_overwrite or do_overwrite):
            if filepath.exists():
                raise FileExistsError(f"Filepath {filepath} already exists!")

        skipped_attrs = {}
        for attr in self._DEL_BEFORE_SAVING_ATTRS:
            if hasattr(self, attr): skipped_attrs[attr] = self.__dict__.pop(attr)

        try:
            with open(filepath, mode='wb') as f:
                match self._PICKLER:
                    case 'dill':
                        if not dill_imported: raise dill_import_error
                        dill.dump(self, f)
                    case 'pickle': pickle.dump(self, f)
                    case _:
                        raise NotImplementedError(
                            f"{self._PICKLER} not supported! Options: {'dill', 'pickle'}"
                        )
        except:
            filepath.unlink()
            raise


        for attr, val in skipped_attrs.items(): setattr(self, attr, val)



from __future__ import annotations

try:
    from tqdm.auto import tqdm

    class TQDMableMixin():
        _SKIP_TQDM_IF_LE = 3

        def __init__(self, *args, **kwargs):
            self.tqdm = kwargs.get('tqdm', tqdm)

        def _tqdm(self, rng, **kwargs):
            if not hasattr(self, 'tqdm'): self.tqdm = tqdm

            if self.tqdm is None: return rng

            try: N = len(rng)
            except: return rng

            if N <= self._SKIP_TQDM_IF_LE: return rng

            return tqdm(rng, **kwargs)
except ImportError as e: pass



import sys, time, numpy as np
sys.path.append('..')

import unittest

from mixins import TimeableMixin

class TimeableDerived(TimeableMixin):
    def __init__(self):
        self.foo = 'foo'
        # Doesn't call super().__init__()! Should still work in this case.

    def uses_contextlib(self, num_seconds: int = 5):
        with self._time_as('using_contextlib'):
            time.sleep(num_seconds)

    @TimeableMixin.TimeAs(key="decorated")
    def decorated_takes_time(self, num_seconds: int = 10):
        time.sleep(num_seconds)

    @TimeableMixin.TimeAs
    def decorated_takes_time_auto_key(self, num_seconds: int = 10):
        time.sleep(num_seconds)

class TestTimeableMixin(unittest.TestCase):
    def test_constructs(self):
        T = TimeableMixin()
        T = TimeableDerived()

    def test_responds_to_methods(self):
        T = TimeableMixin()

        T._register_start('key')
        T._register_end('key')

        T._times_for('key')

        T._register_start('key')
        T._time_so_far('key')
        T._register_end('key')

    def test_pprint_num_unit(self):
        self.assertEqual((5, 'μs'), TimeableMixin._get_pprint_num_unit(5 * 1e-6))

        class Derived(TimeableMixin):
            _CUTOFFS_AND_UNITS = [
                (10, 'foo'),
                (2, 'bar'),
                (None, 'biz')
            ]

        self.assertEqual((3, 'biz'), Derived._get_pprint_num_unit(3, 'biz'))
        self.assertEqual((3, 'foo'), Derived._get_pprint_num_unit(3/20, 'biz'))
        self.assertEqual((1.2, 'biz'), Derived._get_pprint_num_unit(2.4 * 10, 'foo'))

    def test_context_manager(self):
        T = TimeableDerived()

        T.uses_contextlib(num_seconds=1)

        duration = T._times_for('using_contextlib')[-1]
        np.testing.assert_almost_equal(duration, 1, decimal=1)

    def test_times_and_profiling(self):
        T = TimeableDerived()
        T.decorated_takes_time(num_seconds=2)

        duration = T._times_for('decorated')[-1]
        np.testing.assert_almost_equal(duration, 2, decimal=1)

        T.decorated_takes_time_auto_key(num_seconds=2)
        duration = T._times_for('decorated_takes_time_auto_key')[-1]
        np.testing.assert_almost_equal(duration, 2, decimal=1)

        T.decorated_takes_time(num_seconds=1)
        stats = T._duration_stats

        self.assertEqual({'decorated', 'decorated_takes_time_auto_key'}, set(stats.keys()))
        np.testing.assert_almost_equal(1.5, stats['decorated'][0], decimal=1)
        self.assertEqual(2, stats['decorated'][1])
        np.testing.assert_almost_equal(0.5, stats['decorated'][2], decimal=1)
        np.testing.assert_almost_equal(2, stats['decorated_takes_time_auto_key'][0], decimal=1)
        self.assertEqual(1, stats['decorated_takes_time_auto_key'][1])
        self.assertEqual(0, stats['decorated_takes_time_auto_key'][2])

        got_str = T._profile_durations()
        want_str = (
            "decorated_takes_time_auto_key: 2.0 sec\n"
            "decorated:                     1.5 ± 0.5 sec (x2)"
        )
        self.assertEqual(want_str, got_str, msg=f"Want:\n{want_str}\nGot:\n{got_str}")

if __name__ == '__main__': unittest.main()



from __future__ import annotations

import functools, time, numpy as np
from collections import defaultdict
from contextlib import contextmanager
from typing import Optional, Set, Tuple

from .utils import doublewrap

class TimeableMixin():
    _START_TIME = 'start'
    _END_TIME = 'end'

    _CUTOFFS_AND_UNITS = [
        (1000, 'μs'),
        (1000, 'ms'),
        (60, 'sec'),
        (60, 'min'),
        (24, 'hour'),
        (7, 'days'),
        (None, 'weeks')
    ]

    @classmethod
    def _get_pprint_num_unit(cls, x: float, x_unit: str = 'sec') -> Tuple[float, str]:
        x_unit_factor = 1
        for fac, unit in cls._CUTOFFS_AND_UNITS:
            if unit == x_unit: break
            if fac is None:
                raise LookupError(
                    f"Passed unit {x_unit} invalid! "
                    f"Must be one of {', '.join(u for f, u in cls._CUTOFFS_AND_UNITS)}."
                )
            x_unit_factor *= fac

        min_unit = x * x_unit_factor
        upper_bound = 1
        for upper_bound_factor, unit in cls._CUTOFFS_AND_UNITS:
            if (
                (upper_bound_factor is None) or
                (min_unit < upper_bound * upper_bound_factor)
            ): return min_unit / upper_bound, unit
            upper_bound *= upper_bound_factor

    @classmethod
    def _pprint_duration(cls, mean_sec: float, n_times: int = 1, std_seconds: Optional[float] = None) -> str:
        mean_time, mean_unit = cls._get_pprint_num_unit(mean_sec)

        if std_seconds:
            std_time = std_seconds * mean_time/mean_sec
            mean_std_str = f"{mean_time:.1f} ± {std_time:.1f} {mean_unit}"
        else: mean_std_str = f"{mean_time:.1f} {mean_unit}"

        if n_times > 1: return f"{mean_std_str} (x{n_times})"
        else: return mean_std_str

    def __init__(self, *args, **kwargs):
        self._timings = kwargs.get('_timings', defaultdict(list))

    def __assert_key_exists(self, key: str) -> None:
        assert hasattr(self, '_timings') and key in self._timings, f"{key} should exist in self._timings!"

    def _times_for(self, key: str) -> List[float]:
        self.__assert_key_exists(key)
        return [
            t[self._END_TIME] - t[self._START_TIME] for t in self._timings[key] \
                if self._START_TIME in t and self._END_TIME in t
        ]

    def _time_so_far(self, key: str) -> float:
        self.__assert_key_exists(key)
        assert self._END_TIME not in self._timings[key][-1], f"{key} is not currently being timed!"
        return time.time() - self._timings[key][-1][self._START_TIME]

    def _register_start(self, key: str) -> None:
        if not hasattr(self, '_timings'): self._timings = defaultdict(list)

        self._timings[key].append({self._START_TIME: time.time()})

    def _register_end(self, key: str) -> None:
        assert hasattr(self, '_timings')
        assert key in self._timings and len(self._timings[key]) > 0
        assert self._timings[key][-1].get(self._END_TIME, None) is None
        self._timings[key][-1][self._END_TIME] = time.time()

    @contextmanager
    def _time_as(self, key: str):
        self._register_start(key)
        try:
            yield
        finally:
            self._register_end(key)

    @staticmethod
    @doublewrap
    def TimeAs(fn, key: Optional[str] = None):
        if key is None: key = fn.__name__
        @functools.wraps(fn)
        def wrapper_timing(self, *args, seed: Optional[int] = None, **kwargs):
            self._register_start(key=key)
            out = fn(self, *args, **kwargs)
            self._register_end(key=key)
            return out
        return wrapper_timing

    @property
    def _duration_stats(self): 
        out = {}
        for k in self._timings:
            arr = np.array(self._times_for(k))
            out[k] = (arr.mean(), len(arr), arr.std())
        return out

    def _profile_durations(self, only_keys: Optional[Set[str]] = None):
        stats = self._duration_stats

        if only_keys is not None:
            stats = {k: v for k, v in stats.items() if k in only_keys}

        longest_key_length = max(len(k) for k in stats)
        ordered_keys = sorted(stats.keys(), key=lambda k: stats[k][0] * stats[k][1])
        tfk_str = '\n'.join(
            (
                f"{k}:{' '*(longest_key_length - len(k))} "
                f"{self._pprint_duration(*stats[k])}"
            ) for k in ordered_keys
        )
        return tfk_str



from __future__ import annotations

from multiprocessing import Pool

from typing import Callable, Optional, Sequence

class MultiprocessingMixin:
    def __init__(self, *args, multiprocessing_pool_size: Optional[int] = None, **kwargs):
        self.multiprocessing_pool_size = multiprocessing_pool_size

    @property
    def _multiprocessing_pool_size(self):
        if hasattr(self, 'multiprocessing_pool_size'): return self.multiprocessing_pool_size
        else: return None

    @property
    def _use_multiprocessing(self):
        return (self._multiprocessing_pool_size is not None and self._multiprocessing_pool_size > 1)

    def _map(
        self, fn: Callable, iterable: Sequence, tqdm: Optional[Callable] = None, **tqdm_kwargs
    ) -> Sequence:
        if self._use_multiprocessing:
            with Pool(self._multiprocessing_pool_size) as p: 
                if tqdm is None: return p.map(fn, iterable)
                else: return list(tqdm(p.imap(fn, iterable), **tqdm_kwargs))
        elif tqdm is None: return [fn(x) for x in iterable]
        else: return [fn(x) for x in tqdm(iterable, **tqdm_kwargs)]



import sys
sys.path.append('..')

import unittest

try:
    import torch
    from mixins import TensorableMixin

    class TestTensorableMixin(unittest.TestCase):
        def test_constructs(self):
            T = TensorableMixin()

except ImportError: pass

if __name__ == '__main__': unittest.main()




import setuptools

with open("README.md", "r") as fh: long_description = fh.read()

setuptools.setup(
    name="ml_mixins_mmd", # Replace with your own username
    version="0.0.1",
    author="Matthew McDermott",
    author_email="mattmcdermott8@gmail.com",
    description="Various ML / data-science Mixins",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mmcdermott/ML_mixins",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
      'numpy',
    ],
    python_requires='>=3.7',
    test_suite='tests',
)



from __future__ import annotations

try:
    import torch, numpy as np
    from typing import Dict, Hashable, List, Optional, Tuple, Union

    class TensorableMixin():
        Tensorable_T = Union[np.ndarray, List[float], Tuple['Tensorable_T'], Dict[Hashable, 'Tensorable_T']]
        Tensor_T = Union[torch.Tensor, Tuple['Tensor_T'], Dict[Hashable, 'Tensor_T']]

        def __init__(self, *args, **kwargs):
            self.do_cuda = kwargs.get('do_cuda', torch.cuda.is_available)

        def _cuda(self, T: torch.Tensor, do_cuda: Optional[bool] = None):
            if do_cuda is None:
                do_cuda = self.do_cuda if hasattr(self, 'do_cuda') else torch.cuda.is_available

            return T.cuda() if do_cuda else T

        def _from_numpy(self, obj: np.ndarray) -> torch.Tensor:
            # I keep getting errors about "RuntimeError: expected scalar type Float but found Double"
            if obj.dtype == np.float64: obj = obj.astype(np.float32)
            return self._cuda(torch.from_numpy(obj))

        def _nested_to_tensor(self, obj: TensorableMixin.Tensorable_T) -> TensorableMixin.Tensor_T:
            if isinstance(obj, np.ndarray): return self._from_numpy(obj)
            elif isinstance(obj, list): return self._from_numpy(np.array(obj))
            elif isinstance(obj, dict): return {k: self._nested_to_tensor(v) for k, v in obj.items()}
            elif isinstance(obj, tuple): return tuple((self._nested_to_tensor(e) for e in obj))

            raise ValueError(f"Don't know how to convert {type(obj)} object {obj} to tensor!")

except ImportError as e: pass



import sys
sys.path.append('..')

import unittest

from mixins import TQDMableMixin

class TestTQDMableMixin(unittest.TestCase):
    def test_constructs(self):
        T = TQDMableMixin()

if __name__ == '__main__': unittest.main()

